import { useReducer } from 'react';

interface State<T> {
  stack: T[];
}
interface Action<T> {
  type: 'push' | 'pop' | 'clear';
  values?: T[];
}

function createReducer<T>() {
  return function (state: State<T>, action: Action<T>) {
    switch (action.type) {
      case 'push': {
        state.stack.push(...action.values!);
        break;
      }
      case 'pop': {
        state.stack.pop();
        break;
      }
      case 'clear': {
        return { stack: [] };
      }
      default:
        break;
    }
    return {
      stack: state.stack,
    };
  };
}

export function useStack<T>(arr: T[] = []) {
  const [state, dispatch] = useReducer(createReducer<T>(), {
    stack: arr,
  });

  return {
    push: (...values: T[]) => {
      dispatch({ type: 'push', values });
      return values;
    },
    pop: () => {
      const item = state.stack[state.stack.length - 1] || null;
      dispatch({ type: 'pop' });
      return item;
    },
    clear: () => dispatch({ type: 'clear' }),
    forEach: (callback: (value: T, index: number, arr: T[]) => void) =>
      state.stack.reverse().forEach(callback),
    map: (callback: <K>(value: T, index: number, arr: T[]) => K) =>
      state.stack.reverse().map(callback),
    includes: (value: T) => state.stack.includes(value),
    peek: () => state.stack[state.stack.length - 1] || null,
    entries: () => state.stack.entries(),
    values: () => state.stack.values(),
    keys: () => state.stack.keys(),
    size: state.stack.length,
  };
}
